import logging

import eth_event

logger = logging.getLogger(__name__)


class VersionedDict(dict):
    def __init__(self, mapping, version):
        self.v = version
        super().__init__(mapping)

    def __getitem__(self, key):
        v = super().__getitem__(key)
        return v[self.v]


log_signatures = {
    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef": [
        {
            "name": "Transfer",
            "inputs": [
                {"name": "from", "type": "address", "indexed": True},
                {"name": "to", "type": "address", "indexed": True},
                {"name": "value", "type": "uint256", "indexed": False},
            ],
        },
        {
            "name": "Transfer",
            "inputs": [
                {"name": "from", "type": "address", "indexed": True},
                {"name": "to", "type": "address", "indexed": True},
                {"name": "value", "type": "uint256", "indexed": True},
            ],
        },
    ],
    "0xf285329298fd841af46eb83bbe90d1ebe2951c975a65b19a02f965f842ee69c5": [
        {
            "name": "ChangeOwner",
            "inputs": [{"name": "new_owner", "type": "address", "indexed": True}],
        }
    ],
    "0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925": [
        {
            "name": "Approval",
            "inputs": [
                {"name": "owner", "type": "address", "indexed": True},
                {"name": "spender", "type": "address", "indexed": True},
                {"name": "value", "type": "uint256", "indexed": False},
            ],
        },
        {
            "name": "Approval",
            "inputs": [
                {"name": "owner", "type": "address", "indexed": True},
                {"name": "spender", "type": "address", "indexed": True},
                {"name": "value", "type": "uint256", "indexed": True},
            ],
        },
    ],
    "0x42e160154868087d6bfdc0ca23d96a1c1cfa32f1b72ba9ba27b69b98a0d819dc": [
        {
            "name": "AddedBlackList",
            "inputs": [{"name": "user", "type": "address", "indexed": False}],
        },
        {
            "name": "AddedBlackList",
            "inputs": [{"name": "user", "type": "address", "indexed": True}],
        },
    ],
    "0xd7e9ec6e6ecd65492dce6bf513cd6867560d49544421d0783ddf06e76c24470c": [
        {
            "name": "RemovedBlackList",
            "inputs": [{"name": "user", "type": "address", "indexed": False}],
        },
        {
            "name": "RemovedBlackList",
            "inputs": [{"name": "user", "type": "address", "indexed": True}],
        },
    ],
    "0xcb8241adb0c3fdb35b70c24ce35c5eb0c17af7431c99f827d44a445ca624176a": [
        {
            "name": "Issue",
            "inputs": [{"name": "amount", "type": "uint256", "indexed": False}],
        },
        {
            "name": "Issue",
            "inputs": [{"name": "amount", "type": "uint256", "indexed": True}],
        },
    ],
    "0x702d5967f45f6513a38ffc42d6ba9bf230bd40e8f53b16363c7eb4fd2deb9a44": [
        {
            "name": "Redeem",
            "inputs": [{"name": "amount", "type": "uint256", "indexed": False}],
        },
        {
            "name": "Redeem",
            "inputs": [{"name": "amount", "type": "uint256", "indexed": True}],
        },
    ],
    "0xcc358699805e9a8b7f77b522628c7cb9abd07d9efb86b6fb616af1609036a99e": [
        {
            "name": "Deprecate",
            "inputs": [{"name": "newAddress", "type": "address", "indexed": False}],
        },
        {
            "name": "Deprecate",
            "inputs": [{"name": "newAddress", "type": "address", "indexed": True}],
        },
    ],
    "0xb044a1e409eac5c48e5af22d4af52670dd1a99059537a78b31b48c6500a6354e": [
        {
            "name": "Params",
            "inputs": [
                {"name": "feeBasisPoints", "type": "uint256", "indexed": False},
                {"name": "maxFee", "type": "uint256", "indexed": False},
            ],
        }
    ],
    "0x61e6e66b0d6339b2980aecc6ccc0039736791f0ccde9ed512e789a7fbdd698c6": [
        {
            "name": "DestroyedBlackFunds",
            "inputs": [
                {"name": "blackListedUser", "type": "address", "indexed": False},
                {"name": "balance", "type": "uint256", "indexed": False},
            ],
        }
    ],
    "0x9cd6008e8d4ebd34fd9d022278fec7f95d133780ecc1a0dea459fae3e9675390": [
        {
            "name": "TokensSent",
            "inputs": [
                {"name": "amount", "type": "uint256", "indexed": False},
                {"name": "recipient", "type": "bytes32", "indexed": False},
                {"name": "destinationChainId", "type": "uint256", "indexed": False},
                {"name": "receiveToken", "type": "bytes32", "indexed": False},
                {"name": "nonce", "type": "uint256", "indexed": False},
                {"name": "messenger", "type": "uint8", "indexed": False},
            ],
        }
    ],
    "0xe9d840d27ab4032a839c20760fb995af8e3ad1980b9428980ca1c7e072acd87a": [
        {
            "name": "TokensReceived",
            "inputs": [
                {"name": "amount", "type": "uint256", "indexed": False},
                {"name": "recipient", "type": "bytes32", "indexed": False},
                {"name": "nonce", "type": "uint256", "indexed": False},
                {"name": "messenger", "type": "uint8", "indexed": False},
                {"name": "message", "type": "bytes32", "indexed": False},
            ],
        }
    ],
    "0x54791b38f3859327992a1ca0590ad3c0f08feba98d1a4f56ab0dca74d203392a": [
        {
            "name": "MessageSent",
            "inputs": [{"name": "index_topic_1", "type": "bytes32", "indexed": True}],
        }
    ],
    "0xfa628b578e095243f0544bfad9255f49d79d03a5bbf6c85875d05a215e247ad2": [
        {
            "name": "SwapReleased",
            "inputs": [{"name": "encodedSwap", "type": "uint256", "indexed": True}],
        }
    ],
    "0x8d92c805c252261fcfff21ee60740eb8a38922469a7e6ee396976d57c22fc1c9": [
        {
            "name": "SwapExecuted",
            "inputs": [{"name": "encodedSwap", "type": "uint256", "indexed": True}],
        },
    ],
    "0x5ce4019f772fda6cb703b26bce3ec3006eb36b73f1d3a0eb441213317d9f5e9d": [
        {
            "name": "SwapPosted",
            "inputs": [{"name": "encodedSwap", "type": "uint256", "indexed": True}],
        },
    ],
    "0x06724742ccc8c330a39a641ef02a0b419bd09248360680bb38159b0a8c2635d6": [
        {
            "name": "LogDeposit",
            "inputs": [
                {"name": "depositorEthKey", "type": "address", "indexed": False},
                {"name": "starkKey", "type": "uint256", "indexed": False},
                {"name": "vaultId", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "nonQuantizedAmount", "type": "uint256", "indexed": False},
                {"name": "quantizedAmount", "type": "uint256", "indexed": False},
            ],
        },
    ],
    "0x0fcf2162832b2d6033d4d34d2f45a28d9cfee523f1899945bbdd32529cfda67b": [
        {
            "name": "LogNftDeposit",
            "inputs": [
                {"name": "depositorEthKey", "type": "address", "indexed": False},
                {"name": "starkKey", "type": "uint256", "indexed": False},
                {"name": "vaultId", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "tokenId", "type": "uint256", "indexed": False},
                {"name": "assetId", "type": "uint256", "indexed": False},
            ],
        },
    ],
    "0xed94dc026fa9364c53bc0af51cde7f54f3109b3f31fceb26d01396d80e20453b": [
        {
            "name": "LogDepositWithTokenId",
            "inputs": [
                {"name": "depositorEthKey", "type": "address", "indexed": False},
                {"name": "starkKey", "type": "uint256", "indexed": False},
                {"name": "vaultId", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "tokenId", "type": "uint256", "indexed": False},
                {"name": "assetId", "type": "uint256", "indexed": False},
                {"name": "nonQuantizedAmount", "type": "uint256", "indexed": False},
                {"name": "quantizedAmount", "type": "uint256", "indexed": False},
            ],
        },
    ],
    "0xb7477a7b93b2addc5272bbd7ad0986ef1c0d0bd265f26c3dc4bbe42727c2ac0c": [
        {
            "name": "LogWithdrawalPerformed",
            "inputs": [
                {"name": "ownerKey", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "nonQuantizedAmount", "type": "uint256", "indexed": False},
                {"name": "quantizedAmount", "type": "uint256", "indexed": False},
                {"name": "recipient", "type": "address", "indexed": False},
            ],
        },
    ],
    "0xa5cfa8e2199ec5b8ca319288bcab72734207d30569756ee594a74b4df7abbf41": [
        {
            "name": "LogNftWithdrawalPerformed",
            "inputs": [
                {"name": "ownerKey", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "tokenId", "type": "uint256", "indexed": False},
                {"name": "assetId", "type": "uint256", "indexed": False},
                {"name": "recipient", "type": "address", "indexed": False},
            ],
        },
    ],
    "0xc6ba68235f3229e53f3a95cda25543ad54c0f6df2493a06c05fb930bea7966fe": [
        {
            "name": "LogWithdrawalWithTokenIdPerformed",
            "inputs": [
                {"name": "ownerKey", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "tokenId", "type": "uint256", "indexed": False},
                {"name": "assetId", "type": "uint256", "indexed": False},
                {"name": "nonQuantizedAmount", "type": "uint256", "indexed": False},
                {"name": "quantizedAmount", "type": "uint256", "indexed": False},
                {"name": "recipient", "type": "address", "indexed": False},
            ],
        },
    ],
    "0x7e6e15df814c1a309a57686de672b2bedd128eacde35c5370c36d6840d4e9a92": [
        {
            "name": "LogMintWithdrawalPerformed",
            "inputs": [
                {"name": "ownerKey", "type": "uint256", "indexed": False},
                {"name": "assetType", "type": "uint256", "indexed": False},
                {"name": "nonQuantizedAmount", "type": "uint256", "indexed": False},
                {"name": "quantizedAmount", "type": "uint256", "indexed": False},
                {"name": "assetId", "type": "uint256", "indexed": False},
            ],
        },
    ],
    "0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9": [
        {
            "name": "PairCreated",
            "inputs": [
                {"name": "token0", "type": "address", "indexed": True},
                {"name": "token1", "type": "address", "indexed": True},
                {"name": "pair", "type": "address", "indexed": False},
                {"name": "x", "type": "uint256", "indexed": False},
            ],
        },
    ],
    "0x783cca1c0412dd0d695e784568c96da2e9c22ff989357a2e8b1d9b2b4e6b7118": [
        {
            "name": "PoolCreated",
            "inputs": [
                {"name": "token0", "type": "address", "indexed": True},
                {"name": "token1", "type": "address", "indexed": True},
                {"name": "fee", "type": "uint24", "indexed": True},
                {"name": "tickSpacing", "type": "int24", "indexed": False},
                {"name": "pool", "type": "address", "indexed": False},
            ],
        },
    ],
    "0xd78ad95fa46c994b6551d0da85fc275fe613ce37657fb8d5e3d130840159d822": [
        {
            "name": "Swap",
            "inputs": [
                {"name": "sender", "type": "address", "indexed": True},
                {"name": "amount0In", "type": "uint256", "indexed": False},
                {"name": "amount1In", "type": "uint256", "indexed": False},
                {"name": "amount0Out", "type": "uint256", "indexed": False},
                {"name": "amount1Out", "type": "uint256", "indexed": False},
                {"name": "to", "type": "address", "indexed": True},
            ],
        },
    ],
    "0xc42079f94a6350d7e6235f29174924f928cc2ac818eb64fed8004e115fbcca67": [
        {
            "name": "Swap",
            "inputs": [
                {"name": "sender", "type": "address", "indexed": True},
                {"name": "recipient", "type": "address", "indexed": True},
                {"name": "amount0", "type": "int256", "indexed": False},
                {"name": "amount1", "type": "int256", "indexed": False},
                {"name": "sqrtPriceX96", "type": "uint160", "indexed": False},
                {"name": "liquidity", "type": "uint128", "indexed": False},
                {"name": "tick", "type": "int24", "indexed": False},
            ],
        },
    ],
    "0x7fcf532c15f0a6db0bd6d0e038bea71d30d808c7d98cb3bf7268a95bf5081b65": [
        {
            "name": "Withdrawal",
            "inputs": [
                {"name": "src", "type": "address", "indexed": True},
                {"name": "value", "type": "uint256", "indexed": False},
            ],
        },
    ],
}


def is_supported_log(log) -> bool:
    return len(log["topics"]) > 0 and log["topics"][0] in log_signatures


def convert_db_log(db_log) -> dict:
    data_str = db_log.data.hex()
    return {
        "topics": [f"0x{topic.hex()}" for topic in (db_log.topics or [])],
        "data": f"0x{data_str}",
        "address": f"0x{db_log.address.hex()}",
    }


def decoded_log_to_str(decoded_log) -> str:
    name = decoded_log["name"]
    addr = decoded_log["address"].lower()
    params = ",".join([f"{x['name']}={x['value']}" for x in decoded_log["data"]])
    return f"{addr}|{name}({params})".replace("\n", "")


def decode_db_logs(db_logs):
    return [
        x
        for x in [(decode_log(convert_db_log(log)), log) for log in db_logs]
        if x[0] is not None
    ]


def decode_log(log):
    if is_supported_log(log):
        logdef = log_signatures[log["topics"][0]]
        for i in range(0, len(logdef)):
            try:
                return eth_event.decode_log(log, VersionedDict(log_signatures, i))
            except eth_event.EventError as e:
                if i == len(logdef) - 1:
                    logger.info(f"Failed to decode supported log type. {e}. {log}")
    else:
        logger.debug("Can't decode log, not supported yet")
    return None



import importlib
import json

from bitcoinetl.domain.block import BtcBlock
from bitcoinetl.mappers.transaction_mapper import BtcTransactionMapper

from graphsenselib.ingest.utxo import is_coinjoin

from . import resources

mapper = BtcTransactionMapper()
block = BtcBlock()
block.number = 0
block.timestamp = 0
block.hash = 4


def test_c1():
    with importlib.resources.path(
        resources,
        "ab188e2ae90a6e54dad6200f0d2ef188b723d2b393bbd01e65d753058afcba62.json",
    ) as path:
        tx = json.loads(open(path).read())

        converted = mapper.transaction_to_dict(
            mapper.json_dict_to_transaction(tx, block, 0)
        )
        converted["inputs"][0]["addresses"] = ["1EFabZzqMDZALgJmZ7DMHdLj9SPRSuXAvU"]
        converted["inputs"][0]["value"] = 600
        converted["inputs"][1]["addresses"] = ["12JYmnfYU2ghzjwUAspzJsSnmJtK9bZPYR"]
        converted["inputs"][1]["value"] = 5277446

        assert is_coinjoin(converted)


def test_c2():
    with importlib.resources.path(
        resources,
        "59d2780701c8352ad77c026652d67ea596fe4e9316580f02c171e2137fa91578.json",
    ) as path:
        tx = json.loads(open(path).read())

        converted = mapper.transaction_to_dict(
            mapper.json_dict_to_transaction(tx, block, 0)
        )
        converted["inputs"][0]["addresses"] = ["3QxSEAwy5SirDifNnUYtrTmRFAkXcyF9xR"]
        converted["inputs"][0]["value"] = 2700
        converted["inputs"][1]["addresses"] = ["3KSZHyXAPLNmxs98CsAij8ksAENRkEa3zs"]
        converted["inputs"][1]["value"] = 2700
        converted["inputs"][2]["addresses"] = ["1F89hmmrtonJfAQNAqDmeDadcw7AsZcvXG"]
        converted["inputs"][2]["value"] = 200000
        converted["inputs"][3]["addresses"] = ["1F89hmmrtonJfAQNAqDmeDadcw7AsZcvXG"]
        converted["inputs"][3]["value"] = 200000

        assert is_coinjoin(converted)



import logging
import re
from datetime import datetime
from importlib.resources import files, read_text
from typing import Iterable, List, Optional

from parsy import forward_declaration, seq, string

from ..config import config, currency_to_schema_type, keyspace_types
from ..datatypes import BadUserInputError
from ..db import DbFactory
from ..db.cassandra import (
    build_create_stmt,
    build_truncate_stmt,
    normalize_cql_statement,
)
from ..utils import flatten, split_list_on_condition
from ..utils.parsing import (
    anything,
    ci_str_token,
    identifier,
    lexeme,
    space,
    tableidentifier,
)

SCHEMA_TYPE_MAPPING_OVERWRITES = {("account_trx", "transformed"): "account"}

MAGIC_SCHEMA_CONSTANT = "0x8BADF00D"

logger = logging.getLogger(__name__)

comma_sep = ci_str_token(",")
pk = ci_str_token("primary key")
ct = ci_str_token("create table")
ine = ci_str_token("if not exists")
wstmt = ci_str_token("with")
rb = ci_str_token(")")
lb = ci_str_token("(")
rpb = ci_str_token(">")
lpb = ci_str_token("<")
ma = ci_str_token("map")
fs = ci_str_token("frozen")
lst = ci_str_token("list")
ident_lm = lexeme(identifier)
tid = lexeme(tableidentifier)

types = forward_declaration()
types.become(
    (ma + lpb + types + string(",") + space + types + rpb)
    | (lst + lpb + types + rpb)
    | (fs + lpb + types + rpb)
    | ident_lm
)

nested_i_tuple = ident_lm | lb >> ident_lm.sep_by(comma_sep, min=1) << rb
column = seq(
    name=ident_lm,
    type=types,
    pk=(pk.result(True)).optional(),
)
pk_column = seq(
    pk=pk.result(True),
    compound_key=lb >> nested_i_tuple.sep_by(comma_sep) << rb,
)
column_expr = pk_column | column

# TODO: This is likely incomplete and needs some work, with statements support
# is missing
create_parser = seq(
    _select=ct,
    ine=ine.result(True).optional(),
    table=tid,
    columns=lb >> column_expr.sep_by(comma_sep, min=1) << rb,
    withstmt=(wstmt >> anything).optional(),
)


def remove_eol_comments(statement: str) -> str:
    return re.sub(r"\s*?--.*?$", "", statement, 0, re.MULTILINE)


class CreateTableStatement:
    @classmethod
    def from_schema(Cls, schema_str):
        res = create_parser.parse(
            normalize_cql_statement(remove_eol_comments(schema_str))
        )
        keyspace = res["table"]["keyspace"] if "keyspace" in res["table"] else None
        cols = res["columns"]
        w_compound, wo_compound = split_list_on_condition(
            cols, lambda x: "compound_key" in x
        )
        pk_cols = [col for col in wo_compound if col["pk"] is True]
        ld = len(cols) - len(wo_compound)

        if ld == 1:
            pk = w_compound[0]["compound_key"]
        elif ld == 0:
            lpk = len(pk_cols)
            if lpk == 0:
                pk = []
            elif lpk == 1:
                pk = [pk_cols[0]["name"]]
            else:
                raise Exception("Table schema can't have more than one key column.")
        else:
            raise Exception(
                "Table schema can't have more than one compound key column."
            )

        return Cls(
            res["table"]["table"],
            {col["name"]: col["type"] for col in wo_compound},
            pk,
            if_not_exists=res["ine"] is True,
            keyspace=keyspace,
            withstmt=res["withstmt"],
        )

    def __init__(
        self,
        table,
        columns: dict[str, str],
        pks: list,
        if_not_exists=False,
        keyspace=None,
        withstmt=None,
    ):
        self.table = table
        self.columns = columns
        self.pk = pks
        self.if_not_exists = if_not_exists
        self.keyspace = keyspace
        self.with_stmt = withstmt

    def __eq__(self, other):
        return repr(self) == repr(other)

    def __repr__(self):
        return build_create_stmt(
            [f"{n} {t}" for n, t in self.columns.items()],
            [f"({','.join(pk)})" if type(pk) == list else pk for pk in self.pk],
            self.table,
            not self.if_not_exists,
            self.keyspace,
            with_stmt=self.with_stmt,
        )


class Schema:
    def __init__(self, schema_str):
        self.original_schema = schema_str
        self.statements_str = [
            normalize_cql_statement(remove_eol_comments(s))
            for s in schema_str.split(";")
        ]

    def parse_create_table_statements(self) -> Iterable[CreateTableStatement]:
        return [
            CreateTableStatement.from_schema(x)
            for x in self.statements_str
            if x.startswith("create table")
        ]

    def get_schema_string(self, keyspace_name, replication_config):
        replication_factor_magic_constant = (
            MAGIC_SCHEMA_CONSTANT + "_REPLICATION_CONFIG"
        )
        if (
            self.original_schema.count(MAGIC_SCHEMA_CONSTANT) != 3
            or self.original_schema.count(replication_factor_magic_constant) != 1
        ):
            raise Exception(
                f"Expect three occurrences of {MAGIC_SCHEMA_CONSTANT} in schema file,"
                " two for the keyspace_name and one for the replication factor"
                f" config ({replication_factor_magic_constant})."
            )

        return self.original_schema.replace(
            replication_factor_magic_constant, replication_config
        ).replace(MAGIC_SCHEMA_CONSTANT, keyspace_name)


class GraphsenseSchemas:
    RESOUCE_PATH = f"{__package__}.resources"

    def is_extension(filename, schema_type):
        return not any(
            filename.endswith(f"{kst}_{schema_type}_schema.sql")
            for kst in keyspace_types
        )

    def get_schema_files(self):
        return list(files(self.RESOUCE_PATH).iterdir())

    def load_schema_text(self, filename):
        return read_text(self.RESOUCE_PATH, filename)

    def get_by_currency(
        self, currency, keyspace_type=None, no_extensions=False
    ) -> List[tuple[str, str]]:
        schema_type = currency_to_schema_type[currency]
        if (schema_type, keyspace_type) in SCHEMA_TYPE_MAPPING_OVERWRITES:
            schema_type = SCHEMA_TYPE_MAPPING_OVERWRITES[(schema_type, keyspace_type)]
        return [
            x
            for x in self.get_by_schema_type(
                schema_type,
                keyspace_type=keyspace_type,
            )
            if (
                not no_extensions
                or not GraphsenseSchemas.is_extension(x[0], schema_type)
            )
        ]

    def create_keyspaces_if_not_exist(self, env, currency):
        for kstype in keyspace_types:
            self.create_keyspace_if_not_exist(env, currency, kstype)

    def create_keyspace_if_not_exist(self, env, currency, keyspace_type):
        with DbFactory().from_config(env, currency) as db:
            schema = self.get_by_currency(
                currency, keyspace_type=keyspace_type, no_extensions=True
            )
            if len(schema) > 0:
                schema = schema[0][1]
            else:
                raise Exception(
                    "No schema definition found for "
                    f"{env}, {currency}, type: {keyspace_type}"
                )
            keyspacedb = db.by_ks_type(keyspace_type)
            target_ks_name = keyspacedb.keyspace_name()
            if keyspacedb.exists():
                logger.info(
                    f"Keyspace {keyspace_type} for env "
                    f"{env}:{currency} exists: {target_ks_name}, nothing to do"
                )
            else:
                replication_config = (
                    config.get_keyspace_config(env, currency)
                    .keyspace_setup_config[keyspace_type]
                    .replication_config
                )
                schema_to_create = schema.get_schema_string(
                    target_ks_name, replication_config
                )
                db.db().setup_keyspace_using_schema(schema_to_create)
                logger.info(
                    f"Keyspace {keyspace_type} for env "
                    f"{env}:{currency} created on {target_ks_name} "
                    f"with replication config {replication_config}."
                )

                if not keyspacedb.is_configuration_populated():
                    config_defaults = (
                        config.get_keyspace_config(env, currency)
                        .keyspace_setup_config[keyspace_type]
                        .data_configuration
                    )
                    logger.warning(
                        "Config table in transformed not populated."
                        f" Setting default values {config_defaults}."
                    )
                    keyspacedb.ingest("configuration", [config_defaults])

    def create_new_transformed_ks_if_not_exist(
        self, env, currency, suffix=None, no_date=False
    ) -> Optional[str]:
        keyspace_type = "transformed"
        with DbFactory().from_config(env, currency) as db:
            schema = self.get_by_currency(
                currency, keyspace_type=keyspace_type, no_extensions=True
            )
            if len(schema) > 0:
                schema = schema[0][1]
            else:
                raise BadUserInputError(
                    "No schema definition found for "
                    f"{env}, {currency}, type: {keyspace_type}"
                )

            date_str = datetime.now().strftime("%Y%m%d")
            keyspace_name = f"{currency}_transformed"
            if not no_date:
                keyspace_name = f"{keyspace_name}_{date_str}"
            if suffix is not None:
                keyspace_name = f"{keyspace_name}_{suffix}"
            c_db = db.db()
            if c_db.has_keyspace(keyspace_name):
                logger.error(
                    f"Keyspace {keyspace_name} for env "
                    f"{env}:{currency} exists; please remove "
                    "or specify an fresh suffix."
                )
                return None
            else:
                replication_config = (
                    config.get_keyspace_config(env, currency)
                    .keyspace_setup_config[keyspace_type]
                    .replication_config
                )
                schema_to_create = schema.get_schema_string(
                    keyspace_name, replication_config
                )
                c_db.setup_keyspace_using_schema(schema_to_create)
                logger.info(
                    f"New transformed keyspace {keyspace_name} for env "
                    f"{env}:{currency} created "
                    f"with replication config {replication_config}."
                )
                return keyspace_name

    def get_by_schema_type(
        self, schema_type, keyspace_type=None
    ) -> List[tuple[str, str]]:
        if (schema_type, keyspace_type) in SCHEMA_TYPE_MAPPING_OVERWRITES:
            schema_type = SCHEMA_TYPE_MAPPING_OVERWRITES[(schema_type, keyspace_type)]
        return [
            (f, s)
            for (f, s) in self.get_all_schemas()
            if (
                (
                    f.endswith(f"{schema_type}_schema.sql")
                    or f.endswith("generic_schema.sql")
                )
                and (keyspace_type is None or (f.startswith(f"{keyspace_type}_")))
            )
        ]

    def get_all_schemas(self) -> List[tuple[str, Schema]]:
        return [
            (file.name, Schema(self.load_schema_text(file.name)))
            for file in self.get_schema_files()
            if file.name.endswith("sql")
        ]

    def get_db_validation_report(self, env, currency):
        with DbFactory().from_config(env, currency) as db:
            raw_schemas = self.get_by_currency(
                currency, keyspace_type="raw", no_extensions=True
            )
            report_raw = self._validate_against_db_intermal(db.raw, raw_schemas)

            trans_schemas = self.get_by_currency(
                currency, keyspace_type="transformed", no_extensions=True
            )
            report_trans = self._validate_against_db_intermal(
                db.transformed, trans_schemas
            )
        return report_raw + report_trans

    def _validate_against_db_intermal(self, db, schemas):
        report = []
        creates_defs = flatten([s.parse_create_table_statements() for f, s in schemas])
        for x in creates_defs:
            pkeys = []
            ckeys = []
            result = db.get_columns_for_table(x.table)
            expected_cols = x.columns.copy()
            mres = list(result)
            ks_table = f"{db.get_keyspace()}.{x.table}"
            logger.info(f"Validating schema of {ks_table}.")
            if len(mres) == 0:
                report.append(f"Table {ks_table} not present in database.")
            for col in mres:
                assert col.table_name == x.table

                if col.column_name not in expected_cols:
                    report.append(
                        f"Column {ks_table}.{col.column_name} present in "
                        "database not present in schema."
                    )
                else:
                    expected_type = expected_cols[col.column_name]
                    if expected_type != col.type:
                        report.append(
                            f"Expected type {ks_table}{expected_cols} "
                            f"but was {col.type}"
                        )
                    expected_cols.pop(col.column_name)

                if col.kind == "partition_key":
                    pkeys.append((col.column_name, col.position))

                if col.kind == "clustering":
                    ckeys.append((col.column_name, col.position))

            if len(expected_cols) > 0:
                cm = ",".join([f"{k} {t}" for k, t in expected_cols.items()])
                report.append(
                    f"Table {ks_table} Some columns that where expected are "
                    f"missing in the database - {cm}"
                )

            expected_key = x.pk
            pkeys = [n for n, p in sorted(pkeys, key=lambda x: x[1])]
            ckeys = [n for n, p in sorted(ckeys, key=lambda x: x[1])]
            key = (
                ([pkeys[0] if len(pkeys) == 1 else pkeys]) if len(pkeys) > 0 else []
            ) + ckeys
            if expected_key != key:
                report.append(
                    f"The keys of the table {ks_table} does not match "
                    f"the expected keys. Expected: {expected_key} got {key}"
                )

        return report

    def get_table_columns_from_file(self, keyspace_type: str, table: str):
        schemas = self.get_by_schema_type(keyspace_type)[0][1].statements_str
        potential_schemas = [s for s in schemas if table in s]
        assert len(potential_schemas) == 1
        schema_str = potential_schemas[0]
        columns = schema_str.split("(")[1].split(")")[0].split(",")[:-1]
        columns = [c.strip() for c in columns]
        pk_columns = schema_str.split("primary key (")[1].split(")")[0].split(",")
        return columns, pk_columns

    def ensure_table_exists_by_name(
        self, db_transformed, table_name: str, truncate: bool = False
    ):
        keyspace = db_transformed.get_keyspace()
        columns, pk_columns = self.get_table_columns_from_file(keyspace, table_name)
        db_transformed._db.execute(
            build_create_stmt(
                columns,
                pk_columns,
                table_name,
                fail_if_exists=False,
                keyspace=keyspace,
            )
        )
        if truncate:
            db_transformed._db.execute(
                build_truncate_stmt(table_name, keyspace=keyspace)
            )



import logging

from graphsenselib.deltaupdate.update.abstractupdater import (
    TABLE_NAME_DIRTY,
    TABLE_NAME_NEW,
    LegacyUpdateStrategy,
    write_dirty_addresses,
    write_new_addresses,
)

logger = logging.getLogger(__name__)


class UpdateStrategyUtxoLegacy(LegacyUpdateStrategy):
    def __init__(self, db, currency, write_new, write_dirty):
        super().__init__(db, currency, write_new, write_dirty)

    def prepare_database(self):
        super().prepare_database()
        ADDRESS_TABLE_COLUMNS = ["address_prefix text", "address text"]
        ADDRESS_TABLE_PK = ["address_prefix", "address"]

        if self._write_new:
            self._db.transformed.ensure_table_exists(
                TABLE_NAME_NEW,
                ADDRESS_TABLE_COLUMNS
                + ["address_id int", "block_id int", "timestamp int", "tx_hash blob"],
                ADDRESS_TABLE_PK,
                truncate=False,
            )

        if self._write_dirty:
            self._db.transformed.ensure_table_exists(
                TABLE_NAME_DIRTY,
                ADDRESS_TABLE_COLUMNS,
                ADDRESS_TABLE_PK,
                truncate=False,
            )

    def process_batch_impl_hook(self, batch):
        addresses = {}

        for block in batch:
            for row in self._db.raw.get_addresses_in_block(block):
                if row.address is not None and row.address not in addresses:
                    addresses[row.address] = row

        if self._write_dirty:
            write_dirty_addresses(self._db, TABLE_NAME_DIRTY, addresses.keys())

        self._nr_queryied_addresses_batch = len(addresses)
        self._nr_queried_addresses += self._nr_queryied_addresses_batch

        if self._write_new:
            logger.info(
                "Checking existence for "
                f"{self._nr_queryied_addresses_batch} addresses"
            )

            ret = self._db.transformed.known_addresses_batch(
                addresses, inlcude_new_adresses_table=True
            )

            for addr, exists in ret.items():
                # this is very slow but good for crosschecking
                # assert exists == self._db.transformed.knows_address(addr)
                if not exists:
                    self._new_addresses[addr] = addresses[addr]
                    self._nr_new_addresses += 1

            logger.info(
                f"Found {self._nr_new_addresses} new addresses so "
                f"far {(self._nr_new_addresses/self._nr_queried_addresses):.3f} "
                "are new."
            )
            # logging.info(self._db.transformed.knows_address.cache_info())

            # At the end of the batch write data to the db.
            write_new_addresses(
                self._db,
                TABLE_NAME_NEW,
                [
                    (
                        addr,
                        self.consume_address_id(),
                        row.timestamp,
                        row.block_id,
                        row.tx_hash,
                    )
                    for addr, row in self._new_addresses.items()
                ],
            )
            # Resetting here means less memory footprint for this job
            # But more writes to Cassandra.
            self._new_addresses = {}
